#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <time.h>
//llda3 <rows> <cols> <dataT> <#clusters> <wl> <wl_> <out> <prior> 
//rows=rows in dataT and cols=cols in dataT

typedef long int integer;

#define MAXCLUST 100

int main (int argc, char* argv[]) {
  int i, j; float sum=0;
  int rows = atoi(argv[1]); int cols = atoi(argv[2]); int output_dim = atoi(argv[4]);
  float wg = atof(argv[5]); float wg_ = atof(argv[6]); 

  time_t starttime; time_t currenttime;

  gsl_matrix * dataT = gsl_matrix_alloc (rows, cols);
  gsl_matrix * data = gsl_matrix_alloc (cols, rows);
  gsl_matrix * Wg = gsl_matrix_alloc (rows, rows);
  gsl_matrix * Wl = gsl_matrix_alloc (rows, rows);

  gsl_vector * sizes = gsl_vector_alloc(MAXCLUST);
  gsl_vector * v = gsl_vector_alloc(rows);
  gsl_vector * w = gsl_vector_alloc(cols);
  gsl_vector * cl = gsl_vector_alloc(rows);
  gsl_permutation * p = gsl_permutation_alloc(rows);

//  starttime=time(NULL);

/* Inits */
  gsl_matrix_set_zero(Wg); gsl_matrix_set_zero(Wl); gsl_vector_set_all(cl, -1); gsl_vector_set_zero(sizes);

  { FILE* f = fopen(argv[3], "rt"); gsl_matrix_fscanf(f, dataT); fclose (f); }
  for(i=0;i<rows;i++) {gsl_matrix_get_row(w,dataT,i);gsl_matrix_set_col(data,i,w);}
//  for(i=0;i<cols;i++) {for(j=0;j<rows;j++) printf("%.2f ", gsl_matrix_get(data,i,j)); printf("\n");}
  
//argv[8] for feature selection; argv[7] for feature extraction
  if(argv[8] != 0) {
     { 
       FILE* f = fopen(argv[8], "rt"); while(fscanf(f, "%d %d", &i, &j) != EOF) gsl_vector_set(cl, j, i); fclose (f); 
       //FILE* f = fopen(argv[8], "rt"); j = 0; while(fscanf(f, "%d", &i) != EOF) { gsl_vector_set(cl, j, i); j++; } fclose (f);

       for (i=0; i<rows; i++) {
         int clid = gsl_vector_get(cl,i);
         if(clid != -1) { int s = (int) gsl_vector_get(sizes,clid); gsl_vector_set(sizes,clid,s+1); }
       }
     }
//    for (i = 0; i < rows; i++) printf ("%.2f: %d\n", gsl_vector_get(cl,i), i); 
  }

//  starttime=time(NULL);

/* Make Wg and Wl */
  for (i = 0; i < rows; i++)
    for (j = 0; j < rows; j++) {
      if(i != j ) gsl_matrix_set(Wg,i,j,(float)1/rows); else gsl_matrix_set(Wg,i,j,0); 
    }

  for (i = 0; i < rows; i++)
    for (j = 0; j < rows; j++){ 
      if( i == j) gsl_matrix_set(Wl,i,j,0);
      else {
        if(gsl_vector_get(cl, i) != -1 && gsl_vector_get(cl, j) != -1) {
          if(gsl_vector_get(cl, i) == gsl_vector_get(cl, j)) { gsl_matrix_set(Wl,i,j,(float)wg); }
          else { gsl_matrix_set(Wl,i,j,(float)wg_); }
        }
      }
    }

//Wg = Wg - 2Wl
  gsl_matrix_scale(Wl,2);
  gsl_matrix_sub(Wg,Wl); 

/* Make Laplacian of above matrix Wg */
  for (i = 0; i < rows; i++){
    sum = 0;
    for (j = 0; j < rows; j++)
      sum += gsl_matrix_get(Wg,i,j);
    gsl_matrix_set(Wg,i,i,-1*sum);
  }

  gsl_matrix_scale(Wg,-1); 

//  currenttime=time(NULL);
//  printf("time to make Wg, Wl, and Laplacians=%.5f\n", difftime(currenttime, starttime));

if(cols <= rows) {
//  printf("doing EVD approach\n");
/* Begin of LLDA -- eigenvector approach */ 
  gsl_matrix_scale(Wg,(float)1/rows);
  gsl_matrix * T1 = gsl_matrix_alloc (cols, rows);
  gsl_matrix * T2 = gsl_matrix_alloc (cols, cols);
//  starttime=time(NULL);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, data, Wg, 0.0, T1);
//  currenttime=time(NULL);
//  printf("time to make T1=data*Wg is %.5f\n", difftime(currenttime, starttime));

/*
  //modification to use cblas matrix mult libraries
  int M=rows; int N=cols; int K_=cols; double alpha=1;
  int LDA=rows; int LDB=cols; double beta=0; int LDC=rows;
  double* A = (double*) malloc(cols*rows*sizeof(double));
  double* B = (double*) malloc(cols*rows*sizeof(double));
  for(j=0;j<rows;j++) for(i=0;i<cols;i++) A[j*cols + i] = gsl_matrix_get(data, i, j);
  for(j=0;j<cols;j++) for(i=0;i<rows;i++) B[j*rows + i] = gsl_matrix_get(Wg, i, j);
  cblas_dgemm (CblasColMajor, CblasNoTrans, CblasNoTrans, M, N, K_, alpha, A, LDA, B, LDB, beta, C, LDC);
*/

//  starttime=time(NULL);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, T1, dataT, 0.0, T2);
//  currenttime=time(NULL);
//  printf("time to make T1*dataT is %.5f\n", difftime(currenttime, starttime));

//Set up dsyevd and dsyevr
//  starttime=time(NULL);
  char JOBZ = 'V'; char UPLO = 'U'; integer N = cols; integer LDA = N; integer INFO, LWORK, LIWORK;
  double* W = (double*) malloc(cols*sizeof(double));
  double* A = (double*) malloc(cols*cols*sizeof(double));
  for(j=0;j<cols;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(T2, i, j);

// dsyevd specifics
  LWORK = 2*cols*cols + 10*cols + 1; LIWORK = 10*cols + 3;
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  integer* IWORK = (integer*) malloc(LIWORK*sizeof(integer));
//  dsyev(&JOBZ, &UPLO, &N, A, &LDA, W, WORK, &LWORK, &INFO);
  dsyevd_(&JOBZ, &UPLO, &N, A, &LDA, W, WORK, &LWORK, IWORK, &LIWORK, &INFO);

// dsyevr specifics
//  char RANGE = 'I'; double DDUMMY; integer IL=1; integer IU=output_dim; integer M;
//  double ABSTOL=0.0000001; integer LDZ=N; 
//For rangge='A'
//  integer* ISUPPZ = (integer*) malloc(2*cols*sizeof(integer));
//For range='I'
//  integer* ISUPPZ = (integer*) malloc(2*output_dim*sizeof(integer));
//For range='I'
//  double* Z = (double*) malloc(cols*output_dim*sizeof(double));
//For range='A'
//  double* Z = (double*) malloc(cols*cols*sizeof(double));
//  LWORK = 26*cols; LIWORK = 10*cols;
//  double* WORK = (double*) malloc(LWORK*sizeof(double));
//  integer* IWORK = (integer*) malloc(LIWORK*sizeof(integer));
//  dsyevr_(&JOBZ, &RANGE, &UPLO, &N, A, &LDA, &DDUMMY, &DDUMMY, &IL, &IU, &ABSTOL, &M, W, Z, &LDZ, ISUPPZ, WORK, &LWORK, IWORK, &LIWORK, &INFO);
//  if(INFO != 0) printf("ERROR code %d in CLAPACK eigenvector routine\n", INFO);
//  currenttime=time(NULL);
//  printf("time to call CLAPACK eigenvector routine=%.5f\n", difftime(currenttime, starttime));

//evec is column first (for non-SVD approach)
  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
  gsl_vector *eval = gsl_vector_alloc (cols);

//For dsyevd_
  for(i=0;i<cols;i++) for(j=0;j<cols;j++) gsl_matrix_set(evec,j,cols-i-1,A[i*cols + j]);
  for(i=0;i<cols;i++) gsl_vector_set(eval,cols-i-1,W[i]);

//For dsyevr_ 'A'
//  for(i=0;i<cols;i++) for(j=0;j<cols;j++) gsl_matrix_set(evec,j,cols-i-1,Z[i*cols + j]);
//  for(i=0;i<cols;i++) gsl_vector_set(eval,cols-i-1,W[i]);
//For dsyevr_ 'I'
//  for(i=0;i<output_dim;i++) for(j=0;j<cols;j++) gsl_matrix_set(evec,j,output_dim-i-1,Z[i*output_dim + j]);
//  for(i=0;i<output_dim;i++) gsl_vector_set(eval,output_dim-i-1,W[i]);
//  for(i=0;i<cols;i++) printf("%.5f ", W[i]); printf("\n");

/* Old stuff */
//  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
//  gsl_vector *eval = gsl_vector_alloc (cols);
//  gsl_eigen_symmv (T2, eval, evec, ew);
//  gsl_eigen_symmv_free (ew);
//  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);

//comment next two lines for feature selection; uncomment for feature extraction
//  gsl_matrix * work = gsl_matrix_alloc (rows, cols);
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, evec, 0.0, work);

  {
//comment next two for feature extraction; uncomment for feature selection
    { FILE* f = fopen(argv[7], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", gsl_vector_get(eval,j)); fprintf(f, "\n"); fclose (f);  }
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(evec,j,i)); printf("\n");} printf("\n");
//uncomment for feature extraction; comment for feature selection
//    for(i=0; i<rows; i++){ for (j = 0; j < output_dim; j++) printf("%.25f ", gsl_matrix_get(work,i,j)); printf("\n"); }
//testing code for feature extraction; comment out for feature selection
//    { FILE* f = fopen(argv[8], "w"); 
//      for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) fprintf(f, "%.15f ", gsl_matrix_get(evec,j,i)); fprintf(f, "\n");} fprintf(f, "\n");} 
  }
}
     
else {
/* LLDA - SVD approach */
//First do SVD of data matrix
  char JOBU = 'O'; char JOBVT = 'A'; integer M = cols, N = rows, LDA = cols, LDU = cols, LDVT = rows;
  integer LWORK = 0, INFO;
  if(cols > rows) LWORK = 5*cols ; else LWORK = 5*rows;
  double* A = (double*) malloc(cols*rows*sizeof(double));
  gsl_matrix * A_gsl = gsl_matrix_alloc (cols, rows);
  double* S = (double*) malloc(rows*sizeof(double));
  double* VT = (double*) malloc(rows*rows*sizeof(double));
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  double* Udummy = (double*) malloc(1*sizeof(double));
  for(j=0;j<rows;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(data, i, j);
//  for(j=0;j<rows;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(data_meansub, i, j); (from pca.c)
  dgesvd_(&JOBU, &JOBVT, &M, &N, A, &LDA, S, Udummy, &LDU, VT, &LDVT, WORK, &LWORK, &INFO);
//  printf("done svd\n");
  free(WORK);
//  for(i=0;i<rows;i++) for(j=0;j<cols;j++) gsl_matrix_set(U,j,i,A[i*cols + j]); (from pca.c)
  for(i=0;i<rows;i++) for(j=0;j<cols;j++) gsl_matrix_set(A_gsl,j,i,A[i*cols + j]);
//  printf("done svd2\n");
//Then form \Sigma QT (Lg - 2Ll) Q \Sigma
  gsl_matrix * SigmaVT_ = gsl_matrix_alloc (rows, rows);
  for(i=0;i<rows;i++) for(j=0;j<rows;j++) gsl_matrix_set(SigmaVT_,j,i,VT[i*rows + j] * S[j]);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, SigmaVT_, Wg, 0.0, Wl);
  gsl_matrix_transpose(SigmaVT_);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Wl, SigmaVT_, 0.0, Wg);
//  printf("done forming matrix\n");
//Set up and perform dsyevd 
//  starttime=time(NULL);
  char JOBZ = 'V'; char UPLO = 'U'; N = rows; LDA = N; integer LIWORK;
  double* EIGENVALS = (double*) malloc(rows*sizeof(double));
  for(i=0;i<rows;i++) for(j=0;j<rows;j++) VT[i*rows + j] = (double) gsl_matrix_get(Wg, j, i);
  LWORK = 2*rows*rows + 10*rows + 1; LIWORK = 10*rows + 3;
  WORK = (double*) malloc(LWORK*sizeof(double));
  integer* IWORK = (integer*) malloc(LIWORK*sizeof(integer));
  dsyevd_(&JOBZ, &UPLO, &N, VT, &LDA, EIGENVALS, WORK, &LWORK, IWORK, &LIWORK, &INFO);
//  printf("done with EVD\n");
//eigenvecs are in left_sing_vects_of_data \times VT (from above)
  for(i=0;i<rows;i++) for(j=0;j<rows;j++) gsl_matrix_set(Wg,j,i,VT[i*rows + j]);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, A_gsl, Wg, 0.0, data);
//  printf("done with PV\n");

//For SVD-based LLDA
    { FILE* f = fopen(argv[7], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", EIGENVALS[rows-j-1]); fprintf(f, "\n"); fclose (f); }
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(data,j,rows-i-1)); printf("\n");} printf("\n");

}

  gsl_matrix_free (dataT);
  gsl_matrix_free (data);
  gsl_matrix_free (Wg);
  gsl_matrix_free (Wl);
    
  return 0;
}

/****************************************/
/************* OLD STUFF ****************/
/****************************************/

/* Fast LLDA for low dimensions with CLAPACK */
/*  gsl_matrix * T1 = gsl_matrix_alloc (cols, rows);
  gsl_matrix * T2 = gsl_matrix_alloc (cols, cols);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, data, Wg, 0.0, T1);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, T1, dataT, 0.0, T2);
  char JOBZ = 'V'; char UPLO = 'U'; integer N = cols, LDA = cols;
  integer LWORK = 0, INFO;
  if(cols > rows) LWORK = 5*cols ; else LWORK = 5*rows;
  double* W = (double*) malloc(cols*sizeof(double));
  double* A = (double*) malloc(cols*cols*sizeof(double));
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  for(j=0;j<cols;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(T2, i, j);
  dsyev_(&JOBZ, &UPLO, &N, A, &LDA, W, WORK, &LWORK, &INFO);
  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
  gsl_vector *eval = gsl_vector_alloc (cols);
  for(i=0;i<cols;i++) for(j=0;j<cols;j++) gsl_matrix_set(evec,j,cols-i-1,A[i*cols + j]);
  for(i=0;i<cols;i++) gsl_vector_set(eval,cols-i-1,W[i]);
//comment next two lines for llda2; uncomment for llda
//  gsl_matrix * work = gsl_matrix_alloc (rows, cols);
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, evec, 0.0, work);
  {
//comment next two for llda; uncomment for llda2
    { FILE* f = fopen(argv[6], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", gsl_vector_get(eval,j)); fprintf(f, "\n"); fclose (f);  }     
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(evec,j,i)); printf("\n");} printf("\n");
//uncomment for llda; comment for llda2
//    for(i=0; i<rows; i++){ for (j = 0; j < output_dim; j++) printf("%.15f ", gsl_matrix_get(work,i,j)); printf("\n"); }
//    if(argv[12] != 0) { FILE* f = fopen(argv[12],"w"); for(i=0;i<cols;i++){ for (j=0;j<cols;j++) fprintf(f, "%.15f ", gsl_matrix_get(evec,j,i)); fprintf(f,"\n"); } fclose(f); }     
  } */

/* Fast LLDA with CLAPACK*/
/*  char JOBU = 'O'; char JOBVT = 'A'; integer M = cols, N = rows, LDA = cols, LDU = cols, LDVT = rows;
  integer LWORK = 0, INFO;
  if(cols > rows) LWORK = 2*cols ; else LWORK = 2*rows;
  double* A = (double*) malloc(cols*rows*sizeof(double));
  double* S = (double*) malloc(rows*sizeof(double));
  double* VT_ = (double*) malloc(rows*rows*sizeof(double));
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  double* Udummy = (double*) malloc(1*sizeof(double));
  for(j=0;j<rows;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(data, i, j);
  dgesvd_(&JOBU, &JOBVT, &M, &N, A, &LDA, S, Udummy, &LDU, VT_, &LDVT, WORK, &LWORK, &INFO);
  gsl_matrix* LAMBDA = gsl_matrix_alloc(rows,rows); gsl_matrix_set_all(LAMBDA, 0);
  gsl_matrix* VT = gsl_matrix_alloc(rows,rows);
  gsl_matrix* V = gsl_matrix_alloc(rows,rows);
  gsl_matrix * U = gsl_matrix_alloc (cols, rows);
  gsl_matrix * W = gsl_matrix_alloc (cols, rows);
  gsl_matrix * work = gsl_matrix_alloc (rows, rows);
  for(i=0;i<rows;i++) for(j=0;j<cols;j++) gsl_matrix_set(U,j,i,A[i*cols + j]);
  for(i=0;i<rows;i++) for(j=0;j<rows;j++) gsl_matrix_set(VT,j,i,VT_[i*rows + j]);
  for(i=0;i<rows;i++) for(j=0;j<rows;j++) gsl_matrix_set(V,i,j,gsl_matrix_get(VT,j,i));
  for(i=0;i<rows;i++) gsl_matrix_set(LAMBDA,i,i,S[i]);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, V, LAMBDA, 0.0, work);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, LAMBDA, VT, 0.0, V);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Wg, work, 0.0, LAMBDA);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, V, LAMBDA, 0.0, work);
  gsl_eigen_symmv_workspace * ew = gsl_eigen_symmv_alloc (rows);
  gsl_matrix *evec = gsl_matrix_alloc (rows,rows);
  gsl_vector *eval = gsl_vector_alloc (rows);
  gsl_eigen_symmv (work, eval, evec, ew);
  gsl_eigen_symmv_free (ew); 
  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, U, evec, 0.0, W); //-->this takes a while
//  gsl_matrix * W_ = gsl_matrix_alloc (cols, output_dim);
//  gsl_vector * v_ = gsl_vector_alloc (cols);
//  for(i=0;i<output_dim;i++) { gsl_matrix_get_col(v_, W, i); gsl_matrix_set_col(W_, i, v_); }
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, W_, 0.0, work); 

//uncomment for llda; comment for llda2
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, W, 0.0, work); //-->this takes a LONG time
  {
//comment next two for llda; uncomment for llda2
    { FILE* f = fopen(argv[11], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", gsl_vector_get(eval,j)); fprintf(f, "\n"); fclose (f);  }     
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(evec,j,i)); printf("\n");} printf("\n");
//uncomment for llda; comment for llda2
//    for(i=0; i<rows; i++){ for (j = 0; j < output_dim; j++) printf("%.15f ", gsl_matrix_get(work,i,j)); printf("\n"); }
  } */

