#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
//pca <rows> <cols> <dataT> <#clusters> 
//pca2 <rows> <cols> <dataT> <#clusters> <eigens_out> 
//rows=rows in dataT and cols=cols in dataT

typedef long int integer;

int main (int argc, char* argv[]) {
  int i, j; float sum=0;
  int rows = atoi(argv[1]); int cols = atoi(argv[2]); int output_dim = atoi(argv[4]); 

  gsl_matrix * dataT = gsl_matrix_alloc (rows, cols);
  gsl_matrix * data = gsl_matrix_alloc (cols, rows);
  gsl_matrix * dataT_meansub = gsl_matrix_alloc (rows, cols);
  gsl_matrix * data_meansub = gsl_matrix_alloc (cols, rows);

  gsl_vector * overallmean = gsl_vector_alloc(cols);
  gsl_vector * w = gsl_vector_alloc(cols);

/* Inits */
  { FILE* f = fopen(argv[3], "rt"); gsl_matrix_fscanf(f, dataT); fclose (f); }
  for(i=0;i<rows;i++) {gsl_matrix_get_row(w,dataT,i);gsl_matrix_set_col(data,i,w);}
//  for(i=0;i<cols;i++) {for(j=0;j<rows;j++) printf("%.2f ", gsl_matrix_get(data,i,j)); printf("\n");}

/* Compute means */
  for(i=0; i<rows; i++){
    gsl_vector_view idview = gsl_matrix_column(data,i);
    gsl_vector_add(overallmean,&idview.vector);
  }
  gsl_vector_scale(overallmean, (double)1/rows);

/* Make St */
  gsl_matrix_memcpy(data_meansub, data);
  for(i=0; i<rows; i++){
    gsl_vector_view view = gsl_matrix_column(data_meansub,i);
    gsl_vector_sub(&view.vector,overallmean);
  }
  gsl_matrix_transpose_memcpy(dataT_meansub, data_meansub);

/* PCA for low dimensions */
  gsl_matrix * T1 = gsl_matrix_alloc (cols, cols);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, data_meansub, dataT_meansub, 0.0, T1);
//  gsl_matrix_scale(T1, (double)1/rows);

  gsl_eigen_symmv_workspace * ew = gsl_eigen_symmv_alloc (cols);
  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
  gsl_vector *eval = gsl_vector_alloc (cols);
  gsl_eigen_symmv (T1, eval, evec, ew);
  gsl_eigen_symmv_free (ew); 
//comment next three lines for pca2; uncomment for pca
//  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);
//  gsl_matrix * work = gsl_matrix_alloc (rows, cols);
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, evec, 0.0, work);
  {
//comment next two for pca ; uncomment for pca2
    { FILE* f = fopen(argv[5], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", gsl_vector_get(eval,j)); fprintf(f, "\n"); fclose (f);  }
    for(i=0; i<output_dim; i++) {for (j =0; j< cols; j++) printf("%.15f ", gsl_matrix_get(evec,j,i)); printf("\n"); } printf("\n");
//uncomment for pca
//    for(i=0; i<rows; i++){ for (j = 0; j < output_dim; j++) printf("%.15f ", gsl_matrix_get(work,i,j)); printf("\n"); }
  } 

/* PCA for high dimension */
/*  char JOBU = 'O'; char JOBVT = 'A'; integer M = cols, N = rows, LDA = cols, LDU = cols, LDVT = rows;
  integer LWORK = 0, INFO;
  if(cols > rows) LWORK = 5*cols ; else LWORK = 5*rows;
  double* A = (double*) malloc(cols*rows*sizeof(double));
  double* S = (double*) malloc(rows*sizeof(double));
  double* VT_ = (double*) malloc(rows*rows*sizeof(double));
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  double* Udummy = (double*) malloc(1*sizeof(double));
  for(j=0;j<rows;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(data_meansub, i, j);
  dgesvd_(&JOBU, &JOBVT, &M, &N, A, &LDA, S, Udummy, &LDU, VT_, &LDVT, WORK, &LWORK, &INFO);
  gsl_matrix* LAMBDA = gsl_matrix_alloc(rows,rows); gsl_matrix_set_all(LAMBDA, 0);
  gsl_matrix * U = gsl_matrix_alloc (cols, rows);
  gsl_matrix * W = gsl_matrix_alloc (cols, rows);
  gsl_matrix * work = gsl_matrix_alloc (rows, rows);
  for(i=0;i<rows;i++) for(j=0;j<cols;j++) gsl_matrix_set(U,j,i,A[i*cols + j]);
  for(i=0;i<rows;i++) gsl_matrix_set(LAMBDA,i,i,S[i]);
//comment next line for pca2; uncomment for pca
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, U, 0.0, work); //-->this takes a LONG time
  {
//comment next two for pca; uncomment for pca2
    { FILE* f = fopen(argv[5], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", S[j]*S[j]); fprintf(f, "\n"); fclose (f);  }
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(U,j,i)); printf("\n"); } printf("\n");
//uncomment for pca; comment for pca2
//    for(i=0; i<rows; i++){ for (j = 0; j < output_dim; j++) printf("%.15f ", gsl_matrix_get(work,i,j)); printf("\n"); }
  } 
*/

  gsl_matrix_free (dataT);
  gsl_matrix_free (data);
    
  return 0;
}

/****** OLD STUFF *****/
/* PCA for high dimensions */
/*  gsl_matrix * T1 = gsl_matrix_alloc (rows, rows);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT_meansub, data_meansub, 0.0, T1);
  gsl_matrix_scale(T1, (double)1/rows);
  gsl_eigen_symmv_workspace * ew = gsl_eigen_symmv_alloc (rows);
  gsl_matrix *evec = gsl_matrix_alloc (rows,rows);
  gsl_vector *eval = gsl_vector_alloc (rows);
  gsl_eigen_symmv (T1, eval, evec, ew);
  gsl_eigen_symmv_free (ew); 
  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);
  {
//    printf("eigenvals: "); for (j = 0; j < output_dim; j++) printf("%.8f ", gsl_vector_get(eval,j)); printf("\n");
    for(i=0; i<rows; i++){
      for (j = 0; j < output_dim; j++)
        printf("%.15f ", gsl_matrix_get(evec,i,j)*sqrt(gsl_vector_get(eval,j)));
      printf("\n"); 
    }
  } */


