#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
//a.out <rows> <cols> <dataT> <#clusters> <prior>
//rows=rows in dataT and cols=cols in dataT

#define MAXCLUST 100

int main (int argc, char* argv[]) {
  int i, j, k; float sum=0;
  int rows = atoi(argv[1]); int cols = atoi(argv[2]); int output_dim = atoi(argv[4]); 
  float beta = atof(argv[5]); int total=0;

  gsl_matrix * dataT = gsl_matrix_alloc (rows, cols);
  gsl_matrix * data = gsl_matrix_alloc (cols, rows);
  gsl_matrix * dataT_meansub = gsl_matrix_alloc (rows, cols);
  gsl_matrix * data_meansub = gsl_matrix_alloc (cols, rows);
  gsl_matrix * St = gsl_matrix_alloc (cols, cols);
  gsl_matrix * Sw = gsl_matrix_alloc (cols, cols);
  gsl_matrix * tempM = gsl_matrix_alloc (cols, cols);
  gsl_matrix * means = gsl_matrix_alloc(cols, MAXCLUST);
  gsl_matrix * meansT = gsl_matrix_alloc(MAXCLUST,cols);
  gsl_matrix * clusters = gsl_matrix_alloc(rows, MAXCLUST);

  gsl_vector * overallmean = gsl_vector_alloc(cols);
  gsl_vector * sizes = gsl_vector_alloc(MAXCLUST);
  gsl_vector * v = gsl_vector_alloc(rows);
  gsl_vector * w = gsl_vector_alloc(cols);
  gsl_vector * cl = gsl_vector_alloc(rows);
  gsl_permutation * p = gsl_permutation_alloc(rows);

/* Inits */
  gsl_vector_set_all(cl, -1); 
  gsl_matrix_set_zero(means); gsl_vector_set_zero(sizes);
  gsl_matrix_set_all(clusters, -1); gsl_matrix_set_all(Sw, 0);

  { FILE* f = fopen(argv[3], "rt"); gsl_matrix_fscanf(f, dataT); fclose (f); }
  for(i=0;i<rows;i++) {gsl_matrix_get_row(w,dataT,i);gsl_matrix_set_col(data,i,w);}
//  for(i=0;i<cols;i++) {for(j=0;j<rows;j++) printf("%.2f ", gsl_matrix_get(data,i,j)); printf("\n");}
  
/* ENABLED PRIORS */
  { 
    FILE* f = fopen(argv[6], "rt"); while(fscanf(f, "%d %d", &i, &j) != EOF) gsl_vector_set(cl, j, i); fclose (f); 
//    FILE* f = fopen(argv[6], "rt"); j = 0; while(fscanf(f, "%d", &i) != EOF) { gsl_vector_set(cl, j, i); j++; } fclose (f); 
    for (i=0; i<rows; i++) { 
      int clid = gsl_vector_get(cl,i);
      if(clid != -1) {
        int s = (int) gsl_vector_get(sizes,clid);
        gsl_matrix_set(clusters,s,clid,i);
        gsl_vector_set(sizes,clid,s+1);
        total++;
      }
    }
  } 
//  printf("\n");
//  for(i=0; i<rows; i++){for(j = 0; j < MAXCLUST; j++) printf("%.0f ", gsl_matrix_get(clusters,i,j));printf("\n");} printf("\n");
//  for(i=0; i<MAXCLUST; i++) printf("%.0f ", gsl_vector_get(sizes,i)); printf("\n"); printf("\n");

/* Compute means */
  for(i=0; i<MAXCLUST; i++){
    gsl_vector_view meanview = gsl_matrix_column(means,i);
    for(j = 0; j < gsl_vector_get(sizes,i); j++){
       int id = gsl_matrix_get(clusters,j,i);
       gsl_vector_view idview = gsl_matrix_column(data,id);
       gsl_vector_add(&meanview.vector,&idview.vector);
       gsl_vector_add(overallmean,&idview.vector);
    }
    if(gsl_vector_get(sizes,i)) gsl_vector_scale(&meanview.vector, (double)1/gsl_vector_get(sizes,i));
  }
  gsl_vector_scale(overallmean, (double)1/total);

//  for(i=0;i<MAXCLUST;i++){for(j=0;j<cols;j++) printf("%.5f ", gsl_matrix_get(means,j,i)); printf("\n");} printf("\n");
//  for(j=0;j<cols;j++) printf("%.2f ", gsl_vector_get(overallmean,j)); printf("\n"); printf("\n");

/* Make St */
  gsl_matrix_memcpy(data_meansub, data);
  for(i=0; i<rows; i++){
    if(gsl_vector_get(cl,i) != -1) { 
      gsl_vector_view view = gsl_matrix_column(data_meansub,i); 
      gsl_vector_sub(&view.vector,overallmean);
    }
  }
  gsl_matrix_transpose_memcpy(dataT_meansub, data_meansub);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, data_meansub, dataT_meansub, 0.0, St);

  gsl_matrix_scale(St,(float)1/total);
//  printf("St:\n");
//  for(i=0;i<cols;i++){for(j=0;j<cols;j++) printf("%.2f ", gsl_matrix_get(St,i,j)); printf("\n");} printf("\n");

/* Make Sw */   
  for(i=0; i<MAXCLUST; i++){
    if(gsl_vector_get(sizes,i) > 0){
      gsl_vector * classmean = gsl_vector_alloc(cols); gsl_vector_set_all(classmean,0);
      gsl_matrix * Ci = gsl_matrix_alloc(cols,gsl_vector_get(sizes,i));
      gsl_matrix * CiT = gsl_matrix_alloc(gsl_vector_get(sizes,i),cols);
      for(j=0;j<gsl_vector_get(sizes,i);j++){
        int id = gsl_matrix_get(clusters,j,i);
        gsl_vector_view idview = gsl_matrix_column(data,id);
        gsl_matrix_set_col(Ci,j,&idview.vector);
        gsl_vector_add(classmean,&idview.vector);
      }
      gsl_vector_scale(classmean, (double)1/gsl_vector_get(sizes,i));
//      for(j=0;j<cols;j++) printf("%.2f ", gsl_vector_get(classmean,j)); printf("\n"); printf("\n");
      for(j = 0; j < gsl_vector_get(sizes,i); j++){
       gsl_vector_view idview = gsl_matrix_column(Ci,j);
       gsl_vector_sub(&idview.vector,classmean);
      }
      gsl_matrix_transpose_memcpy(CiT, Ci);
      gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, Ci, CiT, 0.0, tempM);
      gsl_matrix_add(Sw,tempM);
//      for(k=0;k<cols;k++){for(j=0;j<cols;j++) printf("%.2f ", gsl_matrix_get(Sw,k,j)); printf("\n");} printf("\n");
    }
  }
  
  gsl_matrix_scale(Sw,(float)1/total);
//  printf("Sw:\n");
//  for(k=0;k<cols;k++){for(j=0;j<cols;j++) printf("%.2f ", gsl_matrix_get(Sw,k,j)); printf("\n");} printf("\n");

/* St - 2Sw (same as Sb - Sw) */
  gsl_matrix_scale(Sw,1+beta);
  gsl_matrix_sub(St,Sw); 

/* MMC */
/*
  char JOBZ = 'V'; char UPLO = 'U'; long int N = cols, LDA = cols;
  long int LWORK, INFO, LIWORK;
  LWORK = 2*cols*cols + 10*cols + 1; LIWORK = 10*cols + 3;
  double* W = (double*) malloc(cols*sizeof(double));
  double* A = (double*) malloc(cols*cols*sizeof(double));
  double* WORK = (double*) malloc(LWORK*sizeof(double));
  long int* IWORK = (long int*) malloc(LIWORK*sizeof(long int));
  for(j=0;j<cols;j++) for(i=0;i<cols;i++) A[j*cols + i] = (double) gsl_matrix_get(St, i, j);
//  dsyev_(&JOBZ, &UPLO, &N, A, &LDA, W, WORK, &LWORK, &INFO);
  dsyevd_(&JOBZ, &UPLO, &N, A, &LDA, W, WORK, &LWORK, IWORK, &LIWORK, &INFO);
  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
  gsl_vector *eval = gsl_vector_alloc (cols);
  for(i=0;i<cols;i++) for(j=0;j<cols;j++) gsl_matrix_set(evec,j,cols-i-1,A[i*cols + j]);
  for(i=0;i<cols;i++) gsl_vector_set(eval,cols-i-1,W[i]);
*/

  gsl_eigen_symmv_workspace * ew = gsl_eigen_symmv_alloc (cols);
  gsl_matrix *evec = gsl_matrix_alloc (cols,cols);
  gsl_vector *eval = gsl_vector_alloc (cols);
  gsl_eigen_symmv (St, eval, evec, ew);
  gsl_eigen_symmv_free (ew);
  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_VAL_DESC);

//comment for feature selection; uncomment for extraction
//  gsl_matrix * work = gsl_matrix_alloc (rows, cols);
//  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, dataT, evec, 0.0, work);

//  for(i=0; i<rows; i++){ for (j = 0; j < cols; j++) printf("%.3f ", gsl_matrix_get(work,i,j)); printf("\n"); }
  {
//uncomment for feature selection; comment otherwise
    { FILE* f = fopen(argv[7], "w"); for (j=0; j<output_dim; j++) fprintf(f, "%.15f ", gsl_vector_get(eval,j)); fprintf(f, "\n"); fclose (f);  }
    for(i=0; i<output_dim; i++) {for (j=0; j<cols; j++) printf("%.15f ", gsl_matrix_get(evec,j,i)); printf("\n");} printf("\n");
//uncomment for feature extraction; comment otherwise
//    for(i=0; i<rows; i++){  for (j = 0; j < output_dim; j++) printf("%.25f ", gsl_matrix_get(work,i,j)); printf("\n"); } 
  } 

  gsl_matrix_free (dataT);
  gsl_matrix_free (data);
  gsl_matrix_free (Sw);
    
  return 0;
}

